from dataclasses import dataclass, field
from typing import Type

from edri.api.handlers import HTTPHandler
from edri.dataclass.event import Event
from edri.dataclass.directive import HTMLResponseDirective
from edri.utility.function import format_url


@dataclass
class RedirectResponseDirective(HTMLResponseDirective):
    """Directive for redirecting to another URL, status 302 will be used"""
    location: Type[Event] | str  # The explicitly required parameter
    kwargs: dict = field(init=False, repr=False, default=None)

    def __init__(self, location: Type[Event] | str, /, **kwargs):
        if isinstance(location, type) and issubclass(location, Event):
            extensions = HTTPHandler.event_type_extensions.get(location)
            if "url_original" in extensions:
                location = format_url(extensions["url_original"], **kwargs)
            else:
                location = extensions["url"]
        self.location = location




from dataclasses import dataclass, fields
from enum import Enum
from http import HTTPMethod
from inspect import isclass
from pathlib import Path
from types import NoneType
from typing import Type, get_origin, get_args, Union

from typeguard import check_type

from edri.api.extensions.url_prefix import PrefixBase
from edri.config.constant import ApiType
from edri.dataclass.event import EventHandlingType, _event, Event
from edri.utility.function import camel2snake


class Cookie:
    def __init__(self, name: str):
        self.name = name


class Header:
    def __init__(self, name: str):
        self.name = name.lower()


class Scope:
    def __init__(self, name: str):
        self.name = name.lower()


@dataclass
class ApiEvent:
    url: str
    resource: str
    handling: EventHandlingType
    event: Type[Event]
    exclude: list[ApiType]
    cookies: dict[str, str]
    headers: dict[str, str]
    scope: dict[str, str]
    template: str | None


api_events: list[ApiEvent] = list()
allowed_types = (str, int, float, bool, Path)


def api(cls=None, /, *, init=True, repr=True, eq=True, order=False,
        unsafe_hash=False, frozen=False, match_args=True,
        kw_only=False, slots=False, weakref_slot=None, url=None, resource=None, handling=None, exclude=None, template=None,
        shareable=False):
    def wrapper(cls):
        cookies = {}
        headers = {}
        scope = {}
        url_prefix = url if url else f"/{camel2snake(cls.__name__)}"

        is_event_subclass = False
        prefixes = set()
        for base in reversed(cls.__bases__):
            if is_event_subclass or issubclass(base, Event):
                is_event_subclass = True

            if issubclass(base, PrefixBase):
                prefixes.add(base)
                url_prefix = base.prefix + url_prefix
        if not is_event_subclass:
            cls = type(cls.__name__, (Event, *tuple(base for base in cls.__bases__ if base not in prefixes)), dict(cls.__dict__))
        else:
            cls = type(cls.__name__, tuple(base for base in cls.__bases__ if base not in prefixes), dict(cls.__dict__))
        cls.__annotations__.pop('method', None)
        dataclass = _event(cls, init=init, repr=repr, eq=eq, order=order, unsafe_hash=unsafe_hash,
                           frozen=frozen, match_args=match_args, kw_only=kw_only, slots=slots,
                           weakref_slot=weakref_slot, shareable=shareable)

        for name, field in ((f.name, f) for f in fields(dataclass)):
            if name.startswith("_"):
                continue
            if name == "response" or name == "method":
                pass
            elif isinstance(field.default, Cookie):
                cookies[name] = field.default.name
            elif isinstance(field.default, Header):
                headers[name] = field.default.name
            elif isinstance(field.default, Scope):
                scope[name] = field.default.name
            elif field.type not in allowed_types and (isclass(field.type) and not issubclass(field.type, Enum)):
                item_type = get_origin(field.type)
                item_args = get_args(field.type)
                if item_type == Union:
                    position = 0 if item_args.index(NoneType) == 1 else 1
                    item_type = item_args[position]
                    if item_type in (list, tuple):
                        item_args = item_type.__args__
                if item_type in (list, tuple):
                    if len(item_args) > 1:
                        raise TypeError("Only one child type is allowed got %s" % len(item_args))
                    elif item_args[0] not in allowed_types and not hasattr(item_args[0], "fromisoformat"):
                        raise TypeError("%s cannot be used as a type for API event" % item_args[0])
                elif item_type not in allowed_types and not hasattr(field.type, "fromisoformat"):
                    raise TypeError("%s cannot be used as a type for API event" % field.type)

        if dataclass.method is None:
            if exclude is None or ApiType.REST not in exclude or ApiType.HTML not in exclude:
                raise ValueError("Any method is required")
        else:
            check_type(dataclass.method, HTTPMethod)

        if template is None and (exclude is None or ApiType.HTML not in exclude):
            raise ValueError("For HTML api template has to be specified")

        api_events.append(ApiEvent(
            url_prefix,
            resource if resource else camel2snake(dataclass.__name__).replace("_", "-"),
            handling if handling else EventHandlingType.SPECIFIC,
            dataclass,
            exclude if exclude is not None else [],
            cookies,
            headers,
            scope,
            template))

        return dataclass

    if cls is None:
        return wrapper

    return wrapper(cls)



from dataclasses import dataclass, field
from multiprocessing import Process
from multiprocessing.connection import Connection
from threading import Thread
from typing import Union, Type, Dict

from edri.dataclass.event import Event


@dataclass
class Worker:
    pipe: Connection
    event: Event
    worker: Union[Thread, Process]
    streams: Dict[Type[Event], str] = field(default_factory=dict)



from logging import getLogger
from pickle import dumps, loads
from socket import socket, AF_INET, SOCK_STREAM, SHUT_RDWR, SOL_SOCKET, SO_KEEPALIVE, IPPROTO_TCP, TCP_KEEPIDLE, \
    TCP_KEEPINTVL, TCP_KEEPCNT
from time import sleep
from typing import Type
from uuid import UUID

from edri.config.constant import SWITCH_BYTES_LENGTH, SwitchMessages, SWITCH_MAX_SIZE
from edri.dataclass.event import Event
from edri.config.setting import SWITCH_HOST, SWITCH_PORT, CACHE_TIMEOUT
from edri.events.edri.router import Demands, LastEvents


class Socket:

    def __init__(self) -> None:
        self.logger = getLogger(__name__)

    def reconnect(self) -> None:
        delay = 5
        while True:
            self.logger.debug("Connecting to switch...")
            self.switch_socket = socket(AF_INET, SOCK_STREAM)
            self.switch_socket.setsockopt(SOL_SOCKET, SO_KEEPALIVE, 1)
            self.switch_socket.setsockopt(IPPROTO_TCP, TCP_KEEPIDLE, 5)
            self.switch_socket.setsockopt(IPPROTO_TCP, TCP_KEEPINTVL, 5)
            self.switch_socket.setsockopt(IPPROTO_TCP, TCP_KEEPCNT, 3)
            try:
                self.switch_socket.settimeout(None)
                self.switch_socket.connect((SWITCH_HOST, SWITCH_PORT))
                delay = 5
                self.logger.info("Connected")
                break
            except ConnectionRefusedError as e:
                self.logger.warning("Connection failed - connection was refused", exc_info=e)
                sleep(delay)
            except TimeoutError as e:
                self.logger.warning("Connection failed - connection takes too long", exc_info=e)
                sleep(delay)
            except Exception as e:
                self.logger.warning("Connection failed - unknown error", exc_info=e)
                sleep(delay)
            delay = min(delay + 5, CACHE_TIMEOUT)

    def send_router_id(self, router_id: UUID) -> None:
        self.logger.debug("Sending router id to switch")
        self._send(router_id.bytes)

    def send_demands(self, router_id: UUID, requests: set[Type[Event]], responses: set[Type[Event]]) -> None:
        data = SwitchMessages.NEW_DEMANDS.to_bytes(SWITCH_BYTES_LENGTH, "big", signed=False)
        demands = Demands(
            router_id=router_id,
            requests=requests,
            responses=responses
        )
        self.logger.info("Sending demands to switch: %s", demands)
        local_messages_bytes = dumps(demands)
        data += len(local_messages_bytes).to_bytes(SWITCH_BYTES_LENGTH, "big", signed=False)
        data += local_messages_bytes
        self._send(data)

    def send_router(self, router_id: UUID, event: Event) -> None:
        self.logger.debug("Sending event to switch: %s",  event)
        data = router_id.bytes
        message_bytes = dumps(event)
        data += len(message_bytes).to_bytes(SWITCH_BYTES_LENGTH, "big", signed=False)
        data += message_bytes
        self._send(data)

    def send_request_last_events(self, event: LastEvents) -> None:
        self.logger.info("Sending request for last events to remote routers")
        data = SwitchMessages.LAST_MESSAGES.to_bytes(SWITCH_BYTES_LENGTH, "big", signed=False)
        message_bytes = dumps(event)
        data += len(message_bytes).to_bytes(SWITCH_BYTES_LENGTH, "big", signed=False)
        data += message_bytes
        self._send(data)
        
    def _send(self, data: bytes) -> None:
        if len(data) > SWITCH_MAX_SIZE:
            self.logger.error("Message is too big")
            return
        self.logger.debug("Sending message to switch: %s %s", len(data), data[SWITCH_BYTES_LENGTH:SWITCH_BYTES_LENGTH*2])
        self.switch_socket.send(data)

    def _load(self, amount: int) -> bytes | None:
        data = bytes()
        while len(data) < amount:
            data += self.switch_socket.recv(amount - len(data))
            if not data:
                return None
        return data

    def receive(self) -> Event:
        while True:
            message_size_bytes = self._load(SWITCH_BYTES_LENGTH)
            if not message_size_bytes:
                self.switch_socket.shutdown(SHUT_RDWR)
                self.switch_socket.close()
                self.logger.error("Received no message size")
                raise ConnectionError

            message_size = int.from_bytes(message_size_bytes, "big", signed=False)
            self.logger.debug("Receiving size message: %s", message_size)
            message = self._load(message_size)
            try:
                event: Event = loads(message)
            except ModuleNotFoundError as e:
                self.logger.error("Event file was not found", exc_info=e)
            except AttributeError as e:
                self.logger.error("Event is not the same", exc_info=e)
            except Exception as e:
                self.logger.critical("Unknown error while loading Event", exc_info=e)
            else:
                self.logger.debug("Received: %s", event)
                return event

    def close(self) -> None:
        if hasattr(self, "switch_socket") and self.switch_socket:
            self.switch_socket.shutdown(SHUT_RDWR)
            self.switch_socket.close()



from abc import ABC
from collections import defaultdict
from http import HTTPStatus, HTTPMethod
from http.cookies import SimpleCookie, Morsel
from io import BytesIO
from json import loads, JSONDecodeError
from mimetypes import guess_type
from multipart import MultipartParser
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Callable, Type, Pattern, Any, Unpack, TypedDict, NotRequired, AnyStr, Generic, TypeVar, BinaryIO
from re import compile
from urllib.parse import parse_qs, unquote, parse_qsl

from edri.api import Headers
from edri.api.dataclass import ReceivedFile
from edri.api.dataclass.api_event import api_events
from edri.api.handlers import BaseHandler
from edri.api.handlers.base_handler import BaseDirectiveHandlerDict
from edri.config.constant import ApiType
from edri.config.setting import MAX_BODY_SIZE, ASSETS_PATH, UPLOAD_FILES_PREFIX
from edri.dataclass.directive import HTTPResponseDirective, ResponseDirective
from edri.dataclass.directive.base import InternalServerErrorResponseDirective
from edri.dataclass.directive.http import CookieResponseDirective, AccessDeniedResponseDirective
from edri.dataclass.event import Event
from edri.utility import NormalizedDefaultDict
from edri.utility.function import camel2snake



class EventTypesExtensionsDict(TypedDict):
    url: str
    url_re: NotRequired[Pattern[AnyStr]]
    url_original: NotRequired[str]
    method: HTTPMethod
    cookies: dict[str, str]
    headers: dict[str, str]
    scope: dict[str, str]
    template: str | None


type EventTypesExtensions = dict[Type[Event], EventTypesExtensionsDict]


class ResponseKW(TypedDict):
    headers: Headers


class ResponseErrorKW(TypedDict):
    headers: NotRequired[Headers]


T = TypeVar('T', bound=HTTPResponseDirective)


class HTTPDirectiveHandlerDict(Generic[T], BaseDirectiveHandlerDict):
    routine: NotRequired[Callable[[T, Headers], None] | Callable[[T], None]]
    status: NotRequired[HTTPStatus]
    headers: NotRequired[list[str]]


class HTTPHandler(BaseHandler, ABC):
    _directive_handlers: dict[Type[T], HTTPDirectiveHandlerDict[T]] = {
        CookieResponseDirective: {
            "routine": lambda directive, headers: (
                {"cookie": HTTPHandler._create_cookie(directive, headers)}
            ),
            "headers": ["cookie"],
        },
        AccessDeniedResponseDirective: {
            "status": HTTPStatus.FORBIDDEN,
        },
        InternalServerErrorResponseDirective: {
            "status": HTTPStatus.INTERNAL_SERVER_ERROR,
        },
    }

    _event_types = None
    _event_type_extensions = None
    _event_type_names = None

    def __init__(self,
                 scope: dict,
                 receive: Callable,
                 send: Callable,
                 headers: Headers):
        super().__init__(scope, receive, send)
        self.headers = headers
        self.cookies = self.parse_cookies(self.headers.get("cookie", ""))
        self.parameters: dict[str, Any] = self.parse_url_parameters()
        self.body = BytesIO()

    @classmethod
    @property
    def api_type(cls) -> ApiType:
        pass

    @classmethod
    @property
    def event_types(cls):
        if cls._event_types is None:
            cls._event_types, cls._event_type_extensions, cls._event_type_names = cls.sort_events(cls.api_type)
        return cls._event_types

    @classmethod
    @property
    def event_type_extensions(cls) -> EventTypesExtensions:
        if cls._event_type_extensions is None:
            cls._event_types, cls._event_type_extensions, cls._event_type_names = cls.sort_events(cls.api_type)
        return cls._event_type_extensions

    @classmethod
    @property
    def event_type_names(cls) -> dict[str, Type[Event]]:
        if cls._event_type_names is None:
            cls._event_types, cls._event_type_extensions, cls._event_type_names = cls.sort_events(cls.api_type)
        return cls._event_type_names

    @staticmethod
    def _create_cookie(directive: CookieResponseDirective, headers: Headers) -> SimpleCookie:
        """
        Creates a SimpleCookie object using the attributes from the CookieResponseDirective.

        Args:
            directive (CookieResponseDirective): The directive containing cookie attributes.
            headers (Dict[str, Any]): The current headers, potentially containing existing cookies.

        Returns:
            SimpleCookie: A SimpleCookie object with the cookie set according to the directive.
        """

        # Initialize the SimpleCookie object with existing cookies, if any
        if headers["cookie"]:
            cookie = headers["cookie"]
            cookie[directive.name] = directive.value
        else:
            cookie = SimpleCookie({directive.name: directive.value})

        # Set additional attributes for the specified cookie name
        if directive.path:
            cookie[directive.name]['path'] = directive.path
        if directive.expires:
            cookie[directive.name]['expires'] = directive.expires.strftime('%a, %d-%b-%Y %H:%M:%S GMT')
        if directive.comment:
            cookie[directive.name]['comment'] = directive.comment
        if directive.domain:
            cookie[directive.name]['domain'] = directive.domain
        if directive.max_age is not None:
            cookie[directive.name]['max-age'] = directive.max_age
        if directive.secure:
            cookie[directive.name]['secure'] = 'True'
        if directive.version is not None:
            cookie[directive.name]['version'] = directive.version
        if directive.httponly:
            cookie[directive.name]['httponly'] = 'True'
        if directive.samesite:
            cookie[directive.name]['samesite'] = directive.samesite

        return cookie

    @staticmethod
    def sort_events(handler_type: ApiType) -> tuple[
        dict[str, dict[HTTPMethod | str, Type[Event] | Pattern[str]]], dict[Type[Event], dict[str, Any]],
        dict[str, Type[Event]]]:
        """
        Organizes API events by their URL and HTTP method for efficient lookup.

        Returns:
            A dictionary organizing events by URL and method.
        """
        sorted_event_types: dict[str, dict] = defaultdict(dict)
        event_type_extensions: dict[Type[Event], dict[str, Any]] = defaultdict(dict)
        event_names: dict[str, Type[Event]] = {}
        url_replace = compile(r"{([^}]+)}")
        for api_event in api_events:
            if handler_type in api_event.exclude:
                continue
            url_key = ""
            is_first_part = True
            url_re = ""
            for url_part in api_event.url[1:].split("/"):
                position = url_part.find("{")
                if position == -1:
                    url_key += f"/{url_part}"
                    continue
                elif is_first_part and position > 0:
                    url_key += f"/{url_part[:position]}"
                if position == 0:
                    url_re += url_replace.sub(r"(/\\b(?P<\1>[^/]*))?", url_part)
                elif position > 0 and not is_first_part:
                    url_re += url_replace.sub(fr"/{url_part[:position]}(?P<\1>[^/]+)", url_part[position:])
                elif position > 0 and is_first_part:
                    url_re += url_replace.sub(r"(?P<\1>[^/]+)", url_part[position:])
                is_first_part = False
            events = sorted_event_types[url_key]
            if url_re:
                event_type_extensions[api_event.event]["url_re"] = compile(url_key + url_re + "$")
                event_type_extensions[api_event.event]["url_original"] = api_event.url

            if api_event.event.method in events:
                raise ValueError("Method %s was already set.", api_event.event.method)
            events[api_event.event.method] = api_event.event
            event_type_extensions[api_event.event]["url"] = url_key
            event_type_extensions[api_event.event]["method"] = api_event.event.method
            event_type_extensions[api_event.event]["cookies"] = api_event.cookies
            event_type_extensions[api_event.event]["headers"] = api_event.headers
            event_type_extensions[api_event.event]["scope"] = api_event.scope
            event_type_extensions[api_event.event]["template"] = api_event.template
            event_names[api_event.event.__name__] = api_event.event

        return (dict(sorted(sorted_event_types.items(), key=lambda item: len(item[0]), reverse=True)),
                dict(sorted(event_type_extensions.items(), key=lambda item: len(item[1]["url"]), reverse=True)),
                event_names)

    def get_headers_binary(self, headers: Headers) -> list[tuple[bytes, bytes]]:
        response_headers = self.get_default_headers()
        response_headers.update(headers)
        headers_list = []
        for header, values in response_headers.items():
            if header == "cookie" and isinstance(values, SimpleCookie):
                for cookie in values.values():
                    headers_list.append((b"Set-Cookie", cookie.OutputString().encode(errors="replace")))
                continue
            for value in values:
                headers_list.append((header.encode(errors="replace"), value.encode(errors="replace")))
        return headers_list

    def get_default_headers(self) -> Headers:
        response_headers = NormalizedDefaultDict[*Headers](list)
        response_headers["Access-Control-Allow-Headers"].append(
            "Accept,Accept-Language,Content-Language,Content-Type,Filename")
        response_headers["Access-Control-Allow-Origin"].extend(self.headers.get("origin", []))
        return response_headers

    async def response_error(self, status, response: bytes | None = None, *args, **kwargs: Unpack[ResponseErrorKW]):
        headers = kwargs.get("headers", None)
        if headers is None:
            headers = NormalizedDefaultDict[*Headers](list)
        if response is None:
            response = 'No more information'.encode("utf-8")
        if "Content-Type" not in headers:
            headers["Content-Type"].append("text/html")
        await self.response(status, response, headers=headers)

    @staticmethod
    def parse_cookies(cookie: str):
        return SimpleCookie(cookie)

    def parse_url_parameters(self) -> dict[str, ...]:
        url_parameters = parse_qs(unquote(self.scope["query_string"].decode()), keep_blank_values=True)
        return {
            camel2snake(key.strip("[]")): value if key.endswith("[]") else value[-1] for key, value in
            url_parameters.items()
        }

    def handle_multipart(self, data: BinaryIO):
        try:
            content_type = self.headers.get("content-type", [])
            if len(content_type) < 2:
                raise ValueError("Boundary not found in content-type header")
            boundary = content_type[1].split("=")[1] if "=" in content_type[1] else None
            if not boundary:
                raise ValueError("Boundary not found in content-type header")
            multipart = MultipartParser(data, boundary, charset="utf-8")
            found_files = []
            for part in multipart:
                self.logger.warning(part)
                if not (part.filename or part.content_type):
                    self.parameters[part.name] = part.file.read().decode("utf-8", errors="replace")
                else:
                    with NamedTemporaryFile(prefix=UPLOAD_FILES_PREFIX, delete=False) as received_temp_file:
                        received_temp_file.write(part.file.read())
                        received_file = ReceivedFile(
                            file_name=part.filename,
                            mime_type=part.content_type,
                            path=Path(received_temp_file.name),
                        )
                        found_files.append(received_file)
            if len(found_files) == 1:
                # If it's only one file, then don't return list, but only the single file
                self.parameters["path"] = found_files[0]
            elif len(found_files) > 1:
                self.parameters["path"] = found_files
        except Exception as e:
            raise Exception("Unknown error") from e

    def handle_json(self):
        self.body.seek(0, 2)
        if self.body.tell() == 0:
            return
        self.body.seek(0)
        try:
            self.parameters.update({camel2snake(key): value for key, value in loads(self.body.read()).items()})
        except JSONDecodeError as e:
            self.logger.warning("Cannot process json data", exc_info=e)
            raise e
        except Exception as e:
            self.logger.error("Unknown error", exc_info=e)
            raise e

    def handle_url_encoded(self, data: BinaryIO):
        body_parsed = parse_qsl(
            self.body.read().decode(errors="ignore") if self.body else data.read().decode(
                errors="ignore"),
            keep_blank_values=True,
        )
        for key, value in body_parsed:
            if key.endswith("[]"):
                key = key[:-2]
                if key not in self.parameters:
                    self.parameters[key] = [value]
                else:
                    self.parameters[key].append(value)
            else:
                self.parameters[key] = value

    async def parse_body(self):
        if not self.headers["content-type"]:
            return
        with NamedTemporaryFile(prefix=UPLOAD_FILES_PREFIX) as temporary_file:
            while self.body.tell() < MAX_BODY_SIZE:
                body = await self.receive()
                data = body.get("body", None)
                if data is not None:
                    self.body.write(data)
                    if not body["more_body"]:
                        break
                else:
                    break
            else:
                if "multipart/form-data" not in self.headers["content-type"]:
                    raise ValueError("Body too big")
                temporary_file.write(self.body.read())
                self.body.truncate()
                while data := await self.receive():
                    temporary_file.write(data.get("body", None))
                    if not data["more_body"]:
                        break
                temporary_file.flush()
                temporary_file.seek(0)
            self.body.seek(0)

            if "application/json" in self.headers["content-type"]:
                self.handle_json()
            elif "multipart/form-data" in self.headers["content-type"]:
                self.handle_multipart(self.body if self.body else temporary_file)
            elif "application/x-www-form-urlencoded" in self.headers["content-type"]:
                self.handle_url_encoded(self.body if self.body else temporary_file)
            else:
                raise NotImplementedError("Unsupported content type: %s", self.headers["content-type"])

    async def response(self, status: HTTPStatus, data: bytes, *args, **kwargs: Unpack[ResponseKW]):
        headers = kwargs["headers"]
        if headers is None:
            headers = NormalizedDefaultDict(list)
        headers["Content-Length"].append(str(len(data)))
        await self.send({
            'type': 'http.response.start',
            'status': status,
            'headers': self.get_headers_binary(headers),
        })

        await self.send({
            'type': 'http.response.body',
            'body': data,
            'more_body': False
        })

    def response_error_prepare(self, status: HTTPStatus, response: bytes | dict | None) -> dict:
        data = {
            "status_code": status.value,
            "status": status.name,
            "description": status.description,
            "reasons": []
        }

        if isinstance(response, Event):
            directives = response.get_response().get_directives()
            for directive in directives:
                handler = self.directive_handlers.get(directive.__class__, None)
                if not handler:
                    continue
                handler_status = handler.get("status", None)
                if not handler_status:
                    continue
                if handler_status.is_client_error:
                    message = getattr(directive, "message", None)
                    if message:
                        data["reasons"].append({"message": message, "status_code": handler_status.value})
        elif isinstance(response, dict):
            data.update(response)
        return data

    async def response_file(self, file) -> bool:
        asset_file = Path(ASSETS_PATH, file if file[0] != "/" else file[1:])
        if not asset_file.is_file():
            self.logger.warning("File not found", exc_info=True)
            return False
        mime_types = guess_type(asset_file)
        headers = NormalizedDefaultDict[*Headers](list)
        headers["Content-Type"].append(mime_types[0] if mime_types[0] else "text/html")
        headers["Content-Length"].append(str(asset_file.lstat().st_size))
        with asset_file.open("rb") as asset_file_open:
            try:
                await self.send({
                    'type': 'http.response.start',
                    'status': HTTPStatus.OK,
                    'headers': self.get_headers_binary(headers),
                })
                chunk_size = 1024 * 1024
                while data := asset_file_open.read(chunk_size):
                    await self.send({
                        'type': 'http.response.body',
                        'body': data,
                        'more_body': len(data) == chunk_size
                    })
            except Exception as e:
                self.logger.error(e, exc_info=e)
        return True

    def get_event_constructors(self):
        constructors = self.event_types.get(self.scope["path"], [])
        for event, extensions in self.event_type_extensions.items():
            url_re = extensions.get("url_re", None)
            if url_re:
                match = url_re.match(self.scope["path"])
                if match:
                    constructors.append(event)
        return constructors

    def get_event_constructor(self) -> Type[Event] | None:
        event_constructors = self.event_types.get(self.scope["path"], [])
        method = HTTPMethod(self.scope["method"])
        if not event_constructors:
            url_found = False
            for event_constructor, extensions in self.event_type_extensions.items():
                url_re = extensions.get("url_re", None)
                if not url_re:
                    continue
                url = extensions.get("url", None)
                if self.scope["path"].startswith(url):
                    match = url_re.match(self.scope["path"])
                    if match and method != extensions["method"]:
                        url_found = True
                    elif match:
                            self.parameters.update(match.groupdict())
                            return event_constructor
                    else:
                        raise ResourceWarning("Unknown url %s", self.scope["path"])

            else:
                if url_found:
                    return None
                else:
                    raise ResourceWarning("Unknown url %s", self.scope["path"])

        return event_constructors.get(method, None)

    def create_event(self, event_constructor: Type[Event]) -> Event | None:
        event_extension = self.event_type_extensions.get(event_constructor, None)
        if event_extension:
            headers = event_extension.get("headers", None)
            if headers:
                self.parameters.update({k: self.headers.get(v, None) for k, v in headers.items()})
            cookies = event_extension.get("cookies", None)
            if cookies:
                self.parameters.update({k: self.cookies.get(v, None) for k, v in cookies.items()})
            scope = event_extension.get("scope", None)
            if scope:
                self.parameters.update({k: self.scope.get(v, None) for k, v in scope.items()})
        return super().create_event(event_constructor)

    def handle_directives(self, directives: list[ResponseDirective]) -> tuple[HTTPStatus, Headers]:
        status, headers = HTTPStatus.OK, NormalizedDefaultDict[*Headers](list)
        for directive in directives:
            try:
                directive_handler = self.directive_handlers[directive.__class__]
            except KeyError as e:
                self.logger.warning("Unknown directive: %s", directive, exc_info=e)
            else:
                if "status" in directive_handler:
                    status = directive_handler["status"]
                if "routine" in directive_handler:
                    if "headers" in directive_handler:
                        headers_demand = directive_handler["headers"]
                        partial_headers = directive_handler["routine"](directive, {k: headers[k] for k in headers_demand})
                        headers.update(partial_headers)
                    else:
                        partial_headers = directive_handler["routine"](directive)
                        for key, value in partial_headers.items():
                            headers[key].extend(value)

        return status, headers

    def convert_type(self, value: Any, annotation: Type) -> Any:
        if isinstance(value, Morsel):
            if annotation == Morsel:
                return value
            else:
                return super().convert_type(value.value, annotation)

        return super().convert_type(value, annotation)



from typing import TypeVar, Optional, Callable
from collections import defaultdict

K = TypeVar('K', bound=str)
V = TypeVar('V')


class NormalizedDefaultDict(defaultdict[K, V]):
    def __init__(
            self,
            default_factory: Optional[Callable[[], V]] = None,
            /,
            *args,
            normalization: Optional[Callable[[K], str]] = None,
            **kwargs
    ) -> None:
        self._normalization = normalization or self._default_normalization

        if args and isinstance(args[0], dict):
            normalized_dict = {self._normalization(k): v for k, v in args[0].items()}
            args = (normalized_dict, *args[1:])
        super().__init__(default_factory, *args, **kwargs)

    def __setitem__(self, key: K, value: V) -> None:
        key = self._normalization(key)
        super().__setitem__(key, value)

    def __getitem__(self, key: K) -> V:
        key = self._normalization(key)
        return super().__getitem__(key)

    @staticmethod
    def _default_normalization(key: K) -> str:
        return key.lower()

    def update(self, *args, **kwargs) -> None:
        for k, v in dict(*args, **kwargs).items():
            self[k] = v



from abc import ABC, ABCMeta

from edri.dataclass.event import Event


class MiddlewareMeta(ABCMeta):
    def __init__(cls, name, bases, attrs):
        super().__init__(name, bases, attrs)

        # Dynamically create the list of methods from the class being created
        required_methods = []
        for base in cls.__mro__:
            for method_name, method in base.__dict__.items():
                if callable(method) and not method_name.startswith('__'):
                    required_methods.append(method_name)

        # Check if at least one method is implemented in the subclass
        if not any(attr in cls.__dict__ for attr in required_methods):
            raise TypeError(
                f"Class '{cls.__name__}' must implement at least one of the following methods: {", ".join(required_methods)}. "
                "Ensure that your class provides an implementation for at least one of these methods."
            )


class Middleware(ABC, metaclass=MiddlewareMeta):
    """
    Abstract base class for middleware components.
    Middleware classes derived from this must implement at least one of
    'process_request' and 'process_response'.
    """

    def process_request(self, event: Event):
        """
        Method to process the request before it reaches the main application logic.
        Should be implemented in a subclass.

        :param event: The request event to be processed.
        """
        raise NotImplementedError("Must implement process_request or process_response")

    def process_response(self, event: Event):
        """
        Method to process the response before it is sent to the client.
        Should be implemented in a subclass.

        :param event: The response event to be processed.
        """
        raise NotImplementedError("Must implement process_request or process_response")

    @classmethod
    def __getattr__(cls, name):
        prefix = "is_"
        if name.startswith(prefix):
            capability_name = f"process_{name[len(prefix):]}"
            method = cls.__dict__.get(capability_name)
            return callable(method)
        raise AttributeError("type object '%s' has no attribute '%s'", cls.__name__, name)


from enum import Enum
from http import HTTPStatus
from json import dumps, JSONEncoder
from pathlib import Path
from typing import Any, TypeVar, Callable, Unpack, TypedDict

from edri.api import Headers
from edri.api.handlers import HTTPHandler
from edri.api.handlers.http_handler import ResponseErrorKW
from edri.config.constant import ApiType
from edri.dataclass.directive import HTTPResponseDirective
from edri.dataclass.event import Event

T = TypeVar('T', bound=HTTPResponseDirective)


class ResponseKW(TypedDict):
    headers: Headers


class RESTHandler(HTTPHandler):
    _directive_handlers: dict[T, Callable[[T], tuple[int, Headers]]] = {}

    class CustomJSONEncoder(JSONEncoder):
        """
        A custom JSON encoder for converting various data types to JSON-compatible
        formats, including support for Enums, datetime objects, Paths, bytes, and
        more.

        Inherits from JSONEncoder to override the default() method for custom serialization.
        """

        def default(self, data) -> Any:
            # noinspection SpellCheckingInspection
            if hasattr(data, "to_json"):
                return data.to_json()
            elif hasattr(data, "isoformat"):
                return data.isoformat()
            elif isinstance(data, Path):
                return data.as_posix()
            elif isinstance(data, bytes) or isinstance(data, bytearray):
                return data.hex()
            elif isinstance(data, Enum):
                return data.value
            else:
                return super().default(data)

    @classmethod
    @property
    def api_type(cls) -> ApiType:
        return ApiType.REST

    async def response(self, status: HTTPStatus, data: Event | bytes, *args, **kwargs: Unpack[ResponseKW]):
        headers = kwargs["headers"]
        if isinstance(data, Event):
            response = data.get_response()
            if "Content-Type" not in headers:
                headers["Content-Type"].append("application/json")
            try:
                body = dumps(response.as_dict(transform=True, keep_concealed=False), ensure_ascii=False, cls=self.CustomJSONEncoder).encode("utf-8")
            except TypeError as e:
                self.logger.error("Object is not JSON serializable", exc_info=e)
                await self.response_error(HTTPStatus.INTERNAL_SERVER_ERROR, {
                    "reasons": [{
                        "status_code": HTTPStatus.INTERNAL_SERVER_ERROR,
                        "message": "Object is not JSON serializable",
                    }]
                })
                return
        else:
            body = data

        return await super().response(status, body, headers=headers)

    async def response_error(self, status: HTTPStatus, response: Event | dict | None = None, *args, **kwargs: Unpack[ResponseErrorKW]):
        data = self.response_error_prepare(status, response)
        response = dumps(data, ensure_ascii=False, cls=self.CustomJSONEncoder).encode("utf-8")
        return await super().response_error(status, response, *args, **kwargs)



from datetime import time
from queue import Queue as OriginalQueue, Full
from typing import Generic

from typing_extensions import TypeVar

T = TypeVar('T')


class Queue(Generic[T], OriginalQueue[T]):
    def unget(self, item, block: bool = True, timeout: int | None = None) -> None:
        """Put an item into the head of the queue.
        If optional args 'block' is true and 'timeout' is None (the default),
        block if necessary until a free slot is available. If 'timeout' is
        a non-negative number, it blocks at most 'timeout' seconds and raises
        the Full exception if no free slot was available within that time.
        Otherwise ('block' is false), put an item on the queue if a free slot
        is immediately available, else raise the Full exception ('timeout'
        is ignored in that case).
        """
        with self.not_full:
            if self.maxsize > 0:
                if not block:
                    if self._qsize() >= self.maxsize:
                        raise Full
                elif timeout is None:
                    while self._qsize() >= self.maxsize:
                        self.not_full.wait()
                elif timeout < 0:
                    raise ValueError("'timeout' must be a non-negative number")
                else:
                    endtime = time() + timeout
                    while self._qsize() >= self.maxsize:
                        remaining = endtime - time()
                        if remaining <= 0.0:
                            raise Full
                        self.not_full.wait(remaining)
            self._unget(item)
            self.unfinished_tasks += 1
            self.not_empty.notify()

    def _unget(self, item):
        self.queue.appendleft(item)



from enum import IntEnum, unique, auto

SWITCH_BYTES_LENGTH = 16
SWITCH_MAX_SIZE = 0xFFFFFFFF
SWITCH_COMMANDS = (1 << SWITCH_BYTES_LENGTH*8) - 1
SCHEDULER_TIMEOUT_MAX = 2147483  # INT_MAX without last three digits
STREAM_CLOSE_MARK = "#&@"

@unique
class SwitchMessages(IntEnum):
    @staticmethod
    def _generate_next_value_(name, start, count, last_values):
        return last_values[-1] - count

    NEW_DEMANDS = SWITCH_COMMANDS
    LAST_MESSAGES = auto()

@unique
class ApiType(IntEnum):
    WS = 0
    REST = auto()
    HTML = auto()



from logging import getLogger
from multiprocessing.connection import Connection
from typing import Optional, List, Type, Generic, TypeVar
from abc import abstractmethod

from edri.dataclass.event import Event
from edri.config.constant import STREAM_CLOSE_MARK
from edri.events.edri.manager import StreamCreate, StreamMessage, StreamClose
from edri.events.edri.manager.worker_quit import WorkerQuit

T = TypeVar('T', bound=Event)


class Worker(Generic[T]):
    """
    A generic worker class designed to communicate with a manager through events. It supports sending and receiving events,
    creating event streams for continuous communication, and handling the lifecycle of these streams.

    Attributes:
        _name (str): The name of the worker, used for logging purposes.
        _manager_pipe (Connection): The communication pipe to the manager.
        event (EventType): The initial event associated with the worker.
        logger (Logger): Logger instance for logging messages.
        _buffer (List[Event]): A buffer for storing received events that are not yet processed.
        _stream_buffer (Optional[Event]): A buffer for the last received stream event.
        _stream_key (Optional[str]): The unique key identifying the current stream.

    Methods:
        __init__: Initializes a new Worker instance.
        do: Abstract method defining the main logic of the worker; to be implemented by subclasses.
        message_send: Sends a message to the manager.
        message_receive: Receives a message from the manager, blocking if no message is available.
        run: Entry point for the worker's execution, wrapping the `do` method with error handling and cleanup.
        stream_close: Closes the current message stream.
        stream_create: Initiates a stream for sending and receiving messages of a specific type.
        stream_exists: Checks if a message stream currently exists.
        stream_poll: Checks if there are messages available in the stream.
        stream_receive: Receives a message from the stream.
        stream_send: Sends a message through the stream.
        stream_wait: Waits for a message in the stream, with an optional timeout.
    """
    def __init__(self, pipe: Connection, event: T, name: str, *args, **kwargs) -> None:
        """
        Initializes a new Worker instance.

        Parameters:
            pipe (Connection): The communication pipe to the manager.
            event (EventType): The initial event associated with this worker.
            name (str): A name for the worker, used for logging.
        """
        self._buffer: List[Event] = []
        self._manager_pipe = pipe
        self._name = name
        self._stream_buffer: Optional[Event] = None
        self._stream_key: Optional[str] = None
        self.event = event
        self.logger = getLogger(self._name)
        self._buffer: List[Event] = []
        self._stream_buffer: Optional[Event] = None
        self._stream_key: Optional[str] = None
        self.args = args
        self.kwargs = kwargs
        if event._stream:
            self._stream_key = event._stream

    def event_send(self, event: Event) -> None:
        """
        Sends a event to the manager.

        Parameters:
            event (Event): The event to be sent.
        """
        self._manager_pipe.send(event)
        self.logger.debug("Event was sent: %s", event)

    def event_receive(self) -> Event:
        """
        Receives a event from the manager. This method blocks until a event is available.

        Returns:
            Event: The received event.
        """
        if self._buffer:
            return self._buffer.pop(0)
        while True:
            event: Event = self._manager_pipe.recv()
            if isinstance(event, StreamClose):
                if self._stream_key == event._stream:
                    self._stream_key = None
                    self.logger.debug("Stream closed")
                else:
                    self.logger.warning("Wrong key for closing stream: %s", event._stream)
                continue
            else:
                return event

    def stream_create(self, event: Event) -> bool:
        """
        Initiates a event stream for continuous communication with the manager, or potentially other workers,
        using a specific type of event.

        Parameters:
            event (Event): The initial event for which the stream is to be created.

        Returns:
            bool: True if the stream was successfully created; False otherwise.
        """
        stream_create = StreamCreate(event=event)
        self.event_send(stream_create)
        stream_create = self.event_receive()
        if stream_create._response._status != 0:
            return False
        self._stream_key = stream_create._stream
        return True

    @property
    def stream_exists(self) -> bool:
        """
        Checks whether a event stream is currently active for this worker.

        Returns:
            bool: True if a stream exists; False otherwise.
        """
        return self._stream_key is not None

    def stream_poll(self) -> bool:
        """
        Non-blocking check to determine if there are events available in the current stream.

        Returns:
            bool: True if there is at least one event available in the stream; False otherwise.
        """
        if self._stream_buffer:
            return True
        while self._manager_pipe.poll():
            event = self._manager_pipe.recv()
            if event._stream:
                self._stream_buffer = event
                return True
            else:
                self._buffer.append(event)
        return False

    def stream_wait(self, timeout: Optional[int] = None) -> bool:
        """
         Waits for a event to be available in the stream, optionally with a timeout. This method is blocking
         and will pause execution until a event is received or the timeout is reached.

         Parameters:
             timeout (Optional[int]): The maximum time to wait for a event, in seconds. If None, waits indefinitely.

         Returns:
             bool: True if a event became available in the stream; False if the timeout was reached without receiving a event.
         """
        if self._stream_buffer:
            return True
        while self.stream_exists:
            if self._manager_pipe.poll(timeout) is not None:
                event = self._manager_pipe.recv()
                if event._stream is None:
                    self._buffer.append(event)
                elif event._stream.endswith(STREAM_CLOSE_MARK):
                    self.logger.debug("Stream closed on both ends: %s", event._stream[:-len(STREAM_CLOSE_MARK)])
                    self._stream_key = None
                    return False
                else:
                    self._stream_buffer = event
                    return True
        return False

    def stream_receive(self) -> Event:
        """
        Receives a event from the current stream. This method should be called after confirming that a event
        is available via `stream_poll` or `stream_wait`.

        Returns:
            Event: The next event from the stream.

        Raises:
            BlockingIOError: If called when no events are available in the stream.
        """
        if self._stream_buffer:
            event = self._stream_buffer
            self._stream_buffer = None
            return event
        else:
            raise BlockingIOError

    def stream_send(self, event: Event) -> None:
        """
        Sends a event through the current stream to the manager or another worker.

        Parameters:
            event (Event): The event to be sent through the stream.
        """
        stream_event = StreamMessage(event=event)
        self._manager_pipe.send(stream_event)

    def stream_close(self, event: Optional[Type[Event]] = None) -> bool:
        """
        Closes the current event stream, optionally sending a final event as part of the closure process.

        Parameters:
            event (Optional[Type[Event]]): The type of the final event to send before closing the stream, if any.

        Returns:
            bool: True if the stream was successfully closed; False if there was an issue closing the stream.
        """
        self.logger.debug("Closing the stream...")
        if not self._stream_key:
            self.logger.warning("Key was not provided")
            return False
        self.logger.debug("Stream closed: %s", self._stream_key)
        stream_close = StreamClose(event=event)
        stream_close._stream = self._stream_key
        self.event_send(stream_close)
        return True

    @abstractmethod
    def do(self) -> None:
        """
        Abstract method that should be implemented by subclasses to define the worker's main logic.
        """
        pass

    def run(self) -> None:
        """
        The main entry point for the worker's execution. It calls the `do` method and handles any exceptions,
        ensuring proper cleanup and communication with the manager upon termination.
        """
        try:
            self.do()
        except Exception as e:
            self.logger.error("Worker %s was closed unexpectedly", self._name, exc_info=e)
        finally:
            worker_quit = WorkerQuit()
            self.event_send(worker_quit)



from dataclasses import _process_class, dataclass, field, asdict
from enum import IntEnum
from typing import Any

from edri.dataclass.directive import ResponseDirective
from edri.utility.function import snake2camel


class ResponseStatus(IntEnum):
    NONE = -1
    OK = 0
    FAILED = 1


@dataclass
class Response:
    _status: ResponseStatus = field(init=False, default=ResponseStatus.NONE)
    _changed: bool | None = field(init=False, default=None)
    _directives: list[ResponseDirective] = field(init=False, default_factory=list)

    def as_dict(self, /, *, transform: bool, keep_concealed: bool) -> dict[str, Any]:
        self._directives = []
        response = {"_response": self._status}
        response.update(asdict(self))
        if transform:
            response = {snake2camel(key, exceptions=("_",)): value for key, value in response.items() if
                        keep_concealed or not key.startswith("_")}
        else:
            response = {key: value for key, value in response.items() if keep_concealed or not key.startswith("_")}

        return response

    def get_status(self) -> ResponseStatus:
        return self._status

    def set_status(self, status: ResponseStatus) -> None:
        self._status = status

    def __setattr__(self, key: str, value: Any) -> None:
        if self._changed is not None and not self._changed:
            super().__setattr__("_changed", True)
            super().__setattr__("_status", ResponseStatus.OK)
        super().__setattr__(key, value)

    def __post_init__(self) -> None:
        self._directives = []
        self._changed = False  # Have to by the last

    def has_changed(self) -> bool:
        return self._changed

    def add_directive(self, directive: ResponseDirective) -> None:
        if self._changed is not None and not self._changed:
            super().__setattr__("_changed", True)
            super().__setattr__("_status", ResponseStatus.OK)
        self._directives.append(directive)

    def get_directives(self) -> list[ResponseDirective]:
        return self._directives


def _response(cls, init, repr, eq, order, unsafe_hash, frozen, match_args, kw_only, slots, weakref_slot):
    if not any(issubclass(base, Response) for base in cls.__bases__):
        cls = type(cls.__name__, (Response, cls), dict(cls.__dict__))

    for name, value in cls.__annotations__.items():
        attribute = getattr(cls, name, False)
        if attribute:
            setattr(cls, name, field(init=False, default=attribute))
        else:
            setattr(cls, name, field(init=False, default=None))

    dataclass = _process_class(cls, init=init, repr=repr, eq=eq, order=order, unsafe_hash=unsafe_hash, frozen=frozen, match_args=match_args,
                               kw_only=kw_only, slots=slots, weakref_slot=weakref_slot)

    return dataclass


def response(cls=None, /, *, init=True, repr=True, eq=True, order=False, unsafe_hash=False, frozen=False,
             match_args=True, kw_only=False, slots=False, weakref_slot=None):
    def wrapper(cls):
        return _response(cls, init=init, repr=repr, eq=eq, order=order, unsafe_hash=unsafe_hash,
                         frozen=frozen, match_args=match_args, kw_only=kw_only, slots=slots, weakref_slot=weakref_slot)

    if cls is None:
        return wrapper  # pragma: no cover

    return wrapper(cls)



from http import HTTPStatus
from os.path import abspath, dirname
from typing import Callable, Unpack, TypedDict, Type

from jinja2 import Environment, FileSystemLoader, TemplateSyntaxError, TemplateNotFound, TemplateRuntimeError

from edri.api import Headers
from edri.api.handlers import HTTPHandler
from edri.api.extensions.url_extension import URLExtension
from edri.api.handlers.http_handler import ResponseErrorKW, HTTPDirectiveHandlerDict
from edri.config.constant import ApiType
from edri.config.setting import TEMPLATE_PATH
from edri.dataclass.directive import HTTPResponseDirective
from edri.dataclass.directive import HTMLResponseDirective
from edri.dataclass.directive.html import RedirectResponseDirective
from edri.dataclass.event import Event
from edri.utility import NormalizedDefaultDict


class ResponseKW(TypedDict):
    headers: Headers


def redirect_response_directive_handler():
    pass


class HTMLHandler(HTTPHandler):
    _directive_handlers: dict[Type[HTTPResponseDirective], HTTPDirectiveHandlerDict] = {
        RedirectResponseDirective: {
            "status": HTTPStatus.FOUND,
            "routine": lambda directive: {"location": [directive.location]},
        }
    }

    @classmethod
    @property
    def api_type(cls) -> ApiType:
        return ApiType.HTML

    def __init__(self, scope: dict, receive: Callable, send: Callable, headers: NormalizedDefaultDict[*Headers]):
        super().__init__(scope, receive, send, headers)
        list_of_templates_dirs = [TEMPLATE_PATH, dirname(abspath(__file__)) + "/../static_pages"]
        self.environment = Environment(loader=FileSystemLoader(list_of_templates_dirs), extensions=[URLExtension])
        self.directive_type = HTMLResponseDirective

    async def response(self, status: HTTPStatus, data: Event | bytes, *args, **kwargs: Unpack[ResponseKW]):
        headers = kwargs["headers"]
        if isinstance(data, Event):
            try:
                template = self.environment.get_template(self.event_type_extensions[data.__class__]["template"])
            except TemplateNotFound as e:
                self.logger.error("Template was not found", exc_info=e)
                await self.response_error(HTTPStatus.INTERNAL_SERVER_ERROR, {
                    "reasons": [{
                        "status_code": HTTPStatus.INTERNAL_SERVER_ERROR,
                        "message": "Template was not found",
                    }]
                })
                return
            except TemplateSyntaxError as e:
                self.logger.error("Template syntax error", exc_info=e)
                await self.response_error(HTTPStatus.INTERNAL_SERVER_ERROR, {
                    "reasons": [{
                        "status_code": HTTPStatus.INTERNAL_SERVER_ERROR,
                        "message": "Template syntax error",
                    }]
                })
                return

            data = data.get_response().as_dict(transform=False, keep_concealed=True)
            data["scope"] = self.scope
            try:
                body = template.render(data).encode()
            except (TemplateRuntimeError, TemplateSyntaxError) as e:
                self.logger.error("Template could not be rendered", exc_info=e)
                await self.response_error(HTTPStatus.INTERNAL_SERVER_ERROR, {
                    "reasons": [{
                        "status_code": HTTPStatus.INTERNAL_SERVER_ERROR,
                        "message": str(e),
                    }]
                })
                return
            except TemplateNotFound as e:
                self.logger.error("Template was not found", exc_info=e)
                await self.response_error(HTTPStatus.INTERNAL_SERVER_ERROR, {
                    "reasons": [{
                        "status_code": HTTPStatus.INTERNAL_SERVER_ERROR,
                        "message": "Template was not found",
                    }]
                })
                return
            except Exception as e:
                self.logger.error("Unknown error", exc_info=e)
                await self.response_error(HTTPStatus.INTERNAL_SERVER_ERROR, {
                    "reasons": [{
                        "status_code": HTTPStatus.INTERNAL_SERVER_ERROR,
                        "message": "Unknown error",
                    }]
                })
                return
            if headers is None:
                headers = NormalizedDefaultDict[*Headers](list)
            if "Content-Type" not in headers:
                headers["Content-Type"].append("text/html; charset=utf-8")
        else:
            body = data

        return await super().response(status, body, headers=headers)

    async def response_error(self, status: HTTPStatus, event: Event | dict | None = None, *args,
                             **kwargs: Unpack[ResponseErrorKW]):
        data = self.response_error_prepare(status, event)
        if status.is_redirection:
            if status in (HTTPStatus.MOVED_PERMANENTLY, HTTPStatus.FOUND, HTTPStatus.SEE_OTHER):
                directive = event.response._directives[0]
                data["redirect_location"] = directive.location if hasattr(directive, "location") else None
            template = self.environment.get_template("status_300.j2")
        elif status.is_client_error:
            template = self.environment.get_template("status_400.j2")
        else:
            template = self.environment.get_template("status_500.j2")
        event = template.render(data).encode()

        return await super().response_error(status, event, *args, **kwargs)



from jinja2 import nodes, TemplateRuntimeError
from jinja2.ext import Extension

from edri.api.handlers import HTTPHandler
from edri.utility.function import format_url


class URLExtension(Extension):
    tags = {'url'}

    def parse(self, parser):
        # Parse the token and create a call to the _render_url method
        lineno = next(parser.stream).lineno

        # Parse the first argument, which is the endpoint name
        endpoint = parser.parse_expression()

        # Parse the optional keyword arguments
        args = [endpoint]
        kwargs = []
        while parser.stream.current.type != 'block_end':
            token = parser.stream.current
            if token.type == 'name' and parser.stream.look().type == 'assign':
                key = parser.stream.current.value
                parser.stream.skip()  # Skip over the key
                parser.stream.skip()  # Skip over the '='
                value = parser.parse_expression()
                kwargs.append(nodes.Keyword(key, value))
            else:
                args.append(parser.parse_expression())

        return nodes.Output([self.call_method('_render_url', args, kwargs)], lineno=lineno)

    def _render_url(self, endpoint, **kwargs):
        event_type = HTTPHandler.event_type_names.get(endpoint)
        try:
            extensions = HTTPHandler.event_type_extensions[event_type]
        except KeyError:
            raise TemplateRuntimeError("Event %s was not found" % endpoint)
        if "url_original" in extensions:
            try:
                formated_url = format_url(extensions["url_original"], **kwargs)
            except KeyError as e:
                raise TemplateRuntimeError(e)
            return formated_url
        else:
            try:
                return extensions["url"]
            except KeyError:
                raise TemplateRuntimeError("Event %s was not found" % endpoint)




from threading import Thread
from queue import Queue
from socket import SHUT_RDWR
from logging import getLogger
from uuid import UUID

from edri.config.constant import SWITCH_BYTES_LENGTH, SwitchMessages
from edri.switch.connection import Connection
from edri.switch.forwarder import ForwardType


class Receiver(Thread):
    def __init__(self, forwarder_queue: Queue[tuple[UUID, ForwardType, bytes]], router_id: UUID, connection: Connection,
                 connections: dict[UUID, Connection]) -> None:
        super(Receiver, self).__init__()
        self.forwarder_queue = forwarder_queue
        self.connection = connection
        self.connections = connections
        self.router_id = router_id
        self.logger = getLogger(f"{__name__}.{router_id}")

    def run(self) -> None:
        try:
            while True:
                try:
                    message_type_bytes = self.connection.socket.recv(SWITCH_BYTES_LENGTH)
                except ConnectionResetError as e:
                    self.logger.error("Router disconnected", exc_info=e)
                    return
                if not message_type_bytes:
                    self.connection.socket.shutdown(SHUT_RDWR)
                    self.logger.error("Router disconnected")
                    return
                try:
                    message_type = int.from_bytes(message_type_bytes, "big", signed=False)
                except ValueError:
                    self.logger.error("Unknown message from: %s", message_type_bytes)
                    return
                self.logger.debug("Received new message: %s", message_type)
                if message_type == SwitchMessages.NEW_DEMANDS:
                    self.logger.info("New demands")
                    demands_size_bytes = self.connection.socket.recv(SWITCH_BYTES_LENGTH)
                    demands_size = int.from_bytes(demands_size_bytes, "big", signed=False)

                    demands_bytes = b""
                    while len(demands_bytes) < demands_size:
                        demands_bytes += self.connection.socket.recv(demands_size)
                    self.connections[self.router_id].demands = demands_bytes
                    self.forwarder_queue.put((self.router_id, ForwardType.OTHERS, demands_bytes))
                elif message_type == SwitchMessages.LAST_MESSAGES:
                    self.logger.debug("Last messages")
                    last_messages_size_bytes = self.connection.socket.recv(SWITCH_BYTES_LENGTH)
                    demands_size = int.from_bytes(last_messages_size_bytes, "big", signed=False)

                    last_messages_bytes = b""
                    while len(last_messages_bytes) < demands_size:
                        last_messages_bytes += self.connection.socket.recv(demands_size)
                    self.forwarder_queue.put((self.router_id, ForwardType.OTHERS, last_messages_bytes))

                else:
                    try:
                        receiver = UUID(bytes=message_type_bytes)
                    except ValueError as e:
                        self.logger.error("Cannot convert to UUID", exc_info=e)
                        return
                    message_size_bytes = self.connection.socket.recv(SWITCH_BYTES_LENGTH)
                    message_size = int.from_bytes(message_size_bytes, "big", signed=False)
                    self.logger.debug("Received message size: %s %s", receiver,message_size)
                    message = b""
                    while len(message) < message_size:
                        message += self.connection.socket.recv(message_size)

                    self.logger.debug("Received message: %sb", len(message))
                    self.forwarder_queue.put((receiver, ForwardType.SPECIFIC, message))
        except Exception as e:
            self.logger.error("Neznámá chyba", exc_info=e)



from dataclasses import dataclass, field
from datetime import datetime, timedelta
from logging import getLogger
from threading import Thread, Event as EventThreading
from time import sleep
from typing import List, Type, Dict, Optional, Generator
from uuid import UUID

from edri.config.setting import CACHE_TIMEOUT, CACHE_INFO_MESSAGE
from edri.dataclass.event import Event
from edri.dataclass.response import ResponseStatus


@dataclass
class CacheItem:
    event: Event
    time: datetime = field(default_factory=datetime.now)


class Cache:
    """
    A cache for temporarily storing events with functionality to append events,
    find events based on type and other criteria, periodically clean expired events,
    and maintain a list of the most recent events for specific needs.

    Attributes:
        items (List[CacheItem]): A list of CacheItem instances representing the cached events.
        logger (Logger): Logger instance for logging cache operations and statuses.
        cleaner (Thread): A background thread for periodically cleaning the cache of
                          expired items.
        cleaner_stop (EventThreading): An event flag used to signal the cleaner thread
                                       to stop.

    Methods:
        append(event: Event) -> datetime: Adds a new event to the cache.
        find(event_type: Type[Event], request: bool, from_time: Optional[datetime]) -> List[Event]:
            Searches for events in the cache that match the given criteria.
        clean(): Periodically removes expired events from the cache based on the
                 CACHE_TIMEOUT setting.
        last_events() -> Dict[int, str]: Retrieves the last events for specific identifiers.
        events_from(event_id: str) -> List[Event]: Gets all events following a specific event ID.
        quit(): Signals the cleaner thread to stop and waits for it to finish.
    """

    def __init__(self) -> None:
        """
        Initializes the Cache instance, starts the background cleaner thread, and
        sets up the necessary attributes for cache management.
        """
        self.items: List[CacheItem] = []
        self.logger = getLogger(__name__)
        self.cleaner = Thread(target=self.clean, daemon=True)
        self.cleaner_stop = EventThreading()
        self.cleaner.start()

    def __del__(self):
        """
        Ensures that the cleaner thread is properly stopped when the Cache instance is
        destroyed. This method sets the stop event for the cleaner thread to signal it
        to stop its operation, preventing potential resource leaks or dangling threads.
        """
        if self.cleaner_stop:
            self.cleaner_stop.set()

    def append(self, event: Event) -> datetime:
        """
        Adds a new event to the cache and returns the timestamp when the event was added.

        Parameters:
            event (Event): The event to be added to the cache.

        Returns:
            datetime: The timestamp when the event was added to the cache.
        """
        item = CacheItem(event)
        self.items.append(item)
        return item.time

    def find(self, event_type: Type[Event], request: bool, from_time: Optional[datetime]) -> List[Event]:
        """
        Finds events in the cache matching the specified type, request status, and time criteria.

        Parameters:
            event_type (Type[Event]): The type of event to search for.
            request (bool): Whether to search for events with a specific request status.
            from_time (Optional[datetime]): The starting time to search for events from. If None,
                                            all matching events are returned regardless of time.

        Returns:
            List[Event]: A list of events matching the specified criteria.
        """
        items = [item for item in self.items if isinstance(item.event, event_type) and (
                item.event.response is None or item.event.response.get_status() == ResponseStatus.NONE) == request]
        if from_time is not None:
            return [item.event for item in items if item.time >= from_time]
        else:
            return [item.event for item in items]

    def clean(self) -> None:
        """
        Background thread method to clean expired events from the cache based on the CACHE_TIMEOUT.
        """
        last_log_sent = datetime.now()
        delta = timedelta(seconds=CACHE_TIMEOUT)
        while not self.cleaner_stop.is_set():
            now = datetime.now()
            limit = now - delta
            self.items = [item for item in self.items if item.time > limit]
            if now - timedelta(seconds=CACHE_INFO_MESSAGE) >= last_log_sent:
                self.logger.info("Count of cached events: %s", len(self.items))
                last_log_sent = now
            sleep(1)

    def last_events(self) -> dict[UUID, str]:
        """
        Retrieves the last events for specific identifiers, useful for tracking the most recent
        state or activity.

        Returns:
            Dict[int, str]: A dictionary mapping identifiers to their last event keys.
        """
        return {item.event._switch.router_id: item.event._key for item in self.items if item.event._switch}

    def events_from(self, key: str) -> Generator[Event, None, None]:
        """
        Generates all events from the cache following a specific event ID.

        This method allows iterating over all events that occur after a given event,
        identified by its unique key. It is useful for retrieving events sequentially
        after a specific point.

        Parameters:
            key (str): The unique identifier key of the event to start from.

        Yields:
            Event: Events occurring after the specified event.
        """
        found = False
        for item in self.items:
            if found:
                yield item.event
                continue
            if item.event._key == key:
                found = True

    def quit(self) -> None:
        """
        Signals the cleaner thread to stop its operation and waits for it to finish, ensuring
        a graceful shutdown of the cache cleaning process.
        """
        self.cleaner_stop.set()
        self.cleaner.join()



from typing import Iterable, TypeVar, Callable


T = TypeVar("T")

def camel2snake(title: str) -> str:
    """
    Converts a camelCase string to a snake_case string.

    Parameters:
        title (str): The camelCase string to be converted.

    Returns:
        str: The converted snake_case string.

    Examples:
        >>> camel2snake("camelCaseExample")
        'camel_case_example'
        >>> camel2snake("CamelCaseExample")
        'camel_case_example'
    """
    new_title = ""
    underscore = False
    for letter in title:
        if 65 <= ord(letter) <= 90:
            if underscore:
                new_title += '_'
                underscore = False
            new_title += chr(ord(letter) + 32)
        else:
            underscore = True
            new_title += letter
    return new_title


def snake2camel(title: str, /, *, exceptions: tuple[str] | None = None) -> str:
    """
    Converts a snake_case string to a camelCase string, with optional exceptions
    where the conversion should not be applied.

    Parameters:
        title (str): The snake_case string to be converted.
        exceptions (Optional[tuple[str]]): A tuple of strings that, when matched at the
                                      beginning of the title, prevent conversion and
                                      return the title as is. Defaults to None.

    Returns:
        str: The converted camelCase string if no exception is matched; otherwise,
             the original string.

    Examples:
        >>> snake2camel("snake_case_example")
        'snakeCaseExample'
        >>> snake2camel("id_example", exceptions=("id_",))
        'id_example'

    Notes:
        - The first segment of the snake_case string remains lowercase, aligning
          with the camelCase convention.
        - If the title starts with any of the specified exceptions, it is returned
          unchanged, useful for preserving certain identifiers.
    """
    if exceptions is None:
        exceptions = ("id_",)
    if title.startswith(exceptions):
        return title
    title_list = title.split('_')
    title_list = [title_list[0]] + [word[0].upper() + word[1:] for word in title_list][1:]
    return ''.join(title_list)


def partition(pred: Callable[[T], bool], iterable: Iterable) -> tuple[list[T], list[T]]:
    """
    Splits an iterable into two lists based on a predicate function.

    Args:
        pred (Callable[[T], bool]): A function that takes an item and returns a boolean value.
                                    Items for which this function returns True will go into the
                                    first list; items for which it returns False will go into the
                                    second list.
        iterable (Iterable): The iterable to be partitioned.

    Returns:
        tuple[list[T], list[T]]: A tuple containing two lists:
            - The first list contains all items from the iterable for which `pred` returned True.
            - The second list contains all items from the iterable for which `pred` returned False.

    Example:
        >>> partition(lambda x: x % 2 == 0, [1, 2, 3, 4])
        ([2, 4], [1, 3])
    """
    trues = []
    falses = []
    for item in iterable:
        if pred(item):
            trues.append(item)
        else:
            falses.append(item)
    return trues, falses

def format_url(url: str, **kwargs):
    """
    Formats a URL string by replacing placeholders with keyword arguments. If a parameter is missing,
    it is considered optional only if it appears at the end of the URL or if all subsequent placeholders
    are also missing. Otherwise, raises a KeyError.

    Args:
        url (str): The URL template containing placeholders in the format `{placeholder}`.
        **kwargs: Keyword arguments representing the replacement values for the placeholders in the URL.

    Returns:
        str: The formatted URL with all placeholders replaced by their corresponding values.
             If optional parameters are missing, trailing placeholders are removed.

    Raises:
        KeyError: If a required URL parameter is missing, and it is not at the end of the URL,
                  or if other placeholders to the right are provided.

    Examples:
        >>> format_url('https://example.com/{user}/{id}', user='john', id=42)
        'https://example.com/john/42'

        Missing an optional trailing parameter:
        >>> format_url('https://example.com/{user}/{id}', user='john')
        'https://example.com/john'

        Missing a parameter that makes the rest optional:
        >>> format_url('https://example.com/{user}/{id}/{page}', user='john')
        'https://example.com/john'

        Raises KeyError if a non-trailing parameter is missing:
        >>> format_url('https://example.com/{user}/{id}/detail', user='john')
        KeyError: 'Optional url parameter is missing: id'

        Raises KeyError when a middle parameter is missing, but a subsequent parameter is present:
        >>> format_url('https://example.com/{user}/{id}/{page}', user='john', page='home')
        KeyError: 'Optional url parameter is missing: id'
    """
    class Helper(dict):
        def __init__(self, **kwargs):
            super().__init__(kwargs)
            self.missing = []

        def __missing__(self, key):
            self.missing.append(key)
            return key

    kwargs = Helper(**kwargs)
    formated_url = url.format_map(kwargs)
    for missing in reversed(kwargs.missing):
        if formated_url.endswith(f"/{missing}"):
            formated_url = formated_url.removesuffix(f"/{missing}")
        else:
            raise KeyError("Optional url parameter is missing: %s" % missing)
    return formated_url



from multiprocessing.queues import Queue
from multiprocessing.connection import wait, Connection
from typing import Optional, Set, cast
from logging import getLogger

from edri.abstract import ManagerBase
from edri.api import Middleware
from edri.api.dataclass import Client
from edri.api.dataclass.api_event import api_events
from edri.config.constant import ApiType
from edri.dataclass.directive.base import InternalServerErrorResponseDirective
from edri.dataclass.event import Event
from edri.dataclass.response import ResponseStatus
from edri.events.api import group, manage, client
from edri.utility import Storage



class Broker(ManagerBase):
    """
    The ApiBroker class acts as an intermediary between the central router and all API handlers,
    managing client registrations, event dispatching, and communication in a multiprocessing
    environment. It extends ManagerBase to leverage common management functionalities and
    adds capabilities specific to API interactions.

    Attributes:
        api_broker_queue (Queue): A multiprocessing queue used to receive events specifically for the API broker.
        clients (Optional[Storage[Client]]): A storage mechanism for registered clients, allowing for quick access and management.
        db: Placeholder attribute for potential database connections or operations, not implemented by default.
        events (dict): A mapping of event resources to event classes, derived from the global api_events definition.
        middlewares_request (list[Middleware]): List of middlewares applied to incoming requests.
        middlewares_response (list[Middleware]): List of middlewares applied to outgoing responses.

    Methods:
        __init__(router_queue: "Queue[Event]", api_broker_queue: "Queue[Event]", middlewares: list[Middleware]):
            Constructor for the ApiBroker class.
        name (property): Returns the name of the ApiBroker instance.
        send_specific(event: Event): Sends an event to a specific client based on the event's key.
        send_all(event: Event): Broadcasts an event to all clients of a specific API type.
        client_register(event: client.Register): Registers a new client and sends a confirmation event.
        client_unregister(event: Event): Unregisters a client based on the event's key.
        event_register(event: manage.Register): Registers a client for specific events and parameters.
        solve(event: Event): Determines the appropriate method to handle an incoming event.
        _prepare(): Prepares the ApiBroker for operation by initializing storage for clients.
        additional_pipes (property): Returns a set of additional pipes to monitor for incoming events.
        run_resolver(): Main loop for processing incoming events and dispatching them to the appropriate handlers.
        resolve_callback(event: Event, pipe: Connection): Callback method for handling events received from clients.
        _prepare_resolvers(): Prepares resolvers for handling specific types of events.
        resolve(event: Event): Processes an event through the appropriate middlewares and forwards it to its destination.

    Usage:
        This class is instantiated with specific queues and is used within a multiprocessing
        application to manage communications between the router and API handlers. It handles
        client registrations, event dispatching, and utilizes middlewares for request and
        response processing.
    """

    def __init__(self, router_queue: Queue[Event], api_broker_queue: Queue[Event], middlewares: list[Middleware]) -> None:
        """
        Initializes an ApiBroker instance with specific queues for routing and
        API broker events.

        Parameters:
            router_queue (Queue): The queue for receiving routed events from other
            components of the system.
            api_broker_queue (Queue): The queue specifically for the API broker to
            receive events.
        """
        super().__init__(router_queue)
        self.api_broker_queue = api_broker_queue
        self.logger = getLogger(__name__)
        self.clients: Storage[Client]
        self.db = None
        self.events = {event.resource: event.event for event in api_events}
        self.middlewares_request = [middleware for middleware in middlewares if middleware.is_request]
        self.middlewares_response = [middleware for middleware in middlewares if middleware.is_response]

    @property
    def name(self) -> str:
        """
        Property returning the name of the ApiBroker instance. Primarily used for
        logging and identification purposes.

        Returns:
            str: The name "ApiBroker".
        """
        return "ApiBroker"

    def _send(self, client: Client, event: Event) -> None:
        """
        Attempts to send an event to a specific client. If the connection is broken
        or reset during sending, it logs a warning and unregisters the client.

        Parameters:
            client (Client): The client object representing the target for the event.
            event (Event): The event to be sent to the client.

        Returns:
            None
        """
        try:
            client.socket.send(event)
        except (BrokenPipeError, ConnectionResetError) as e:
            self.logger.warning("Cannot be send %s", client.socket, exc_info=e)
            self.client_unregister(event)

    def send_specific(self, event: Event) -> None:
        """
        Sends an event to the specific client identified by the event's key. Logs
        a warning if the client is not found.

        Parameters:
            event (Event): The event containing the key of the target client.

        Returns:
            None
        """
        self.logger.debug("-> %s", event)
        if not event._key:
            self.logger.error("Key not found in the event %s", event)
            return
        try:
            client = self.clients[event._key]
        except KeyError:
            self.logger.warning("Client was not found: %s", event)
            return
        self._send(client, event)

    def send_all(self, event: Event) -> None:
        """
         Broadcasts an event to all clients that are of a specific API type (e.g., WebSocket).
         It iterates through all clients and sends the event to those matching the type.

         Parameters:
             event (Event): The event to be broadcasted.

         Returns:
             None
         """
        self.logger.debug("->* %s", event)
        for client in self.clients.values():
            if client.type == ApiType.WS:
                self._send(client, event)

    ##### API #####
    def client_register(self, event: client.Register) -> None:
        """
        Handles the registration of a new client based on the received register event.
        It assigns a unique key to the client, stores it, and acknowledges the registration.

        Parameters:
            event (client.Register): The event containing the client's registration information.

        Returns:
            None
        """
        event._key = self.clients.append(Client(event.socket, ApiType.WS))
        event.socket = None
        event.response.set_status(ResponseStatus.OK)
        self.send_specific(event)

    def client_unregister(self, event: Event) -> None:
        """
        Handles the unregistration of a client, removing them from the stored clients
        based on the event's key. It performs cleanup if necessary.

        Parameters:
            event (Event): The event containing the key of the client to be unregistered.

        Returns:
            None
        """
        try:
            if event._key:
                client = self.clients[event._key]
                client.socket.close()
                del self.clients[event._key]
            else:
                self.logger.warning("Client cant be unregistered because key is missing!")
        except KeyError as e:
            self.logger.debug("Client was not found!", exc_info=e)

    def event_register(self, event: manage.Register) -> None:
        """
          Registers a client for specific events and their parameters, allowing for
          customized event handling. This method updates the client's event subscriptions
          and their associated parameters based on the provided event registration details.

          Parameters:
              event (manage.Register): An object representing the event registration request.
              This object must include a non-empty `_key` attribute identifying the client,
              a list of `events` the client wishes to subscribe to, and corresponding
              `parameters` and `values` for those events.

          Returns:
              None

          Raises:
              KeyError: If the event `_key` attribute is missing or if the `_key` does not
              correspond to any existing client in the `clients` dictionary.

              ValueError: If the `events` list in the `event` object is empty, indicating
              that no events have been specified for registration.

              LookupError: If any of the events specified in the `event.events` list do not
              exist in the `self.events` mapping, indicating an attempt to register for
              undefined events.

          This method first verifies that the event `_key` is present and corresponds to an
          existing client. It then proceeds to update the client's event subscriptions and
          parameters. If successful, the client's response status is set to `OK`, and the
          updated event registration details are sent back to the client via their socket.
          """
        if not event._key:
            raise AttributeError("Event key is missing!")
        try:
            client = self.clients[event._key]
        except KeyError as e:
            self.logger.error("Client was not found: %s", event, exc_info=e)
            return

        client.events = {self.events[event] for event in event.events}
        client.parameters = dict(zip(event.parameters, event.values))
        event.response.set_status(ResponseStatus.OK)
        client.socket.send(event)

    def solve(self, event: Event) -> None:
        """
        Determines how to handle an incoming event, directing it to specific clients
        or broadcasting it based on the event's properties.

        Parameters:
            event (Event): The event to be handled.

        Returns:
            None
        """
        if event._key:
            self.send_specific(event)
        else:
            found = False
            for client in self.clients.values():
                if event.__class__ in client.events:
                    found = True
                    client.socket.send(event)
            if not found:
                self.send_all(event)

    def _prepare(self) -> None:
        """
         Prepares the ApiBroker for operation by initializing necessary components,
         such as client storage.

         Returns:
             None
         """
        super()._prepare()
        self.clients: Storage[Client] = Storage()

    @property
    def additional_pipes(self) -> Set[Connection]:
        """
        Provides a set of additional pipes that should be monitored for incoming
        events, specifically the API broker queue.

        Returns:
            Set[Pipe]: A set containing the reader end of the API broker queue.
        """
        return {self.api_broker_queue._reader}

    def run_resolver(self) -> None:
        """
        The main event loop of the ApiBroker, continuously checking for and processing
        incoming events from various sources.

        Returns:
            None
        """
        while True:
            pipes: set[Connection] = set()
            if self.router_pipe:
                pipes.add(self.router_pipe)
            if self._workers:
                pipes.update(worker.pipe for worker in self._workers.values())
            pipes.update(self.additional_pipes)
            try:
                active_pipes = cast(list[Connection], wait(pipes, timeout=10))
                for active_pipe in active_pipes:
                    if active_pipe == self.router_pipe:
                        event: Event = self.router_pipe.recv()
                        self.logger.debug("Received event: %s", event)
                        self.resolve(event)
                    else:
                        try:
                            event = active_pipe.recv()
                        except EOFError as e:
                            self.logger.error("Communication problems", exc_info=e)
                            continue
                        except OSError as e:
                            self.logger.error("OS problems", exc_info=e)
                            continue
                        self.resolve_callback(event, active_pipe)

            except KeyboardInterrupt:
                return

    def resolve_callback(self, event: Event, pipe: Connection) -> None:
        """
        Handles incoming events received from clients or other components, dispatching
        them to the appropriate handling method based on their type.

        Parameters:
            event (Event): The incoming event to be handled.
            pipe (Connection): The communication pipe from which the event was received.

        Returns:
            None
        """
        self.logger.debug("<- Client %s", event)
        if isinstance(event, group.Client):
            if isinstance(event, client.Register):
                self.client_register(event)
            elif isinstance(event, client.Unregister):
                self.client_unregister(event)
        elif isinstance(event, group.Manage):
            if isinstance(event, manage.Register):
                self.event_register(event)
        else:
            for middleware in self.middlewares_request:
                try:
                    middleware.process_request(event)
                except Exception as e:
                    self.logger.error("Unknown error while processing request in middleware", exc_info=e)
                    event.response.set_status(ResponseStatus.FAILED)
                    event.response.add_directive(InternalServerErrorResponseDirective(f"Unknown error while processing request in middleware: {middleware.__class__.__name__}"))
                    self.resolve(event)
                    return
                if not isinstance(event, Event):
                    self.logger.error("Wrong type returned by %s", middleware.__class__.__name__)
                    event.response.set_status(ResponseStatus.FAILED)
                    event.response.add_directive(InternalServerErrorResponseDirective(f"Wrong type returned by {middleware.__class__.__name__}"))
                    self.resolve(event)
                    return
                if event.response.get_status() != ResponseStatus.NONE:
                    self.logger.debug("Event %s returned by %s", event, middleware.__class__.__name__)
                    self.resolve(event)
                    return
            self.logger.debug("Router <- API Broker %s", event)
            self.router_queue.put(event)

    def _prepare_resolvers(self) -> None:
        """
        Prepares the resolvers for handling specific types of events, enhancing
        the base implementation with API-specific event handling.

        Returns:
            None
        """
        super()._prepare_resolvers()
        for event in api_events:
            if "response" not in event.event.__annotations__:
                self._requests[event.event] = self.solve
            else:
                self._responses[event.event] = self.solve

    def resolve(self, event: Event) -> None:
        for middleware in self.middlewares_response:
            event = middleware.process_response(event)
            if event.response.get_status() != ResponseStatus.OK:
                self.logger.debug("Event %s returned by %s", event, middleware.__class__.__name__)
                super().resolve(event)
                return
        super().resolve(event)



from collections import defaultdict
from multiprocessing import Queue
from multiprocessing.connection import Connection
from typing import Dict, Type, Optional, Set
from logging import getLogger

from edri.abstract import ManagerBase
from edri.dataclass.event import Event, shareable_events
from edri.dataclass.response import ResponseStatus
from edri.events.edri import router, group
from edri.events.edri.router import LastEvents
from edri.router import Cache, HealthChecker


class Router:
    """
    Manages routing of events between components within the system, handling subscriptions
    to events, distributing events to subscribed components, and managing health checks.

    Attributes:
        queue (Queue[Event]): A queue for receiving events to be routed.
        logger (Logger): Logger instance for logging routing operations and outcomes.
        connections (Dict[str, Connection]): A dictionary mapping component names to their connections.
        subscribed_requests (Dict[Type[Event], Set[Connection]]): A dictionary mapping event types to sets
                                                                   of connections subscribed to those event types for requests.
        subscribed_responses (Dict[Type[Event], Set[Connection]]): Similar to subscribed_requests, but for responses.
        connector_pipe (Connection): A special connection for the connector component, if any.
        cache (Cache): An instance of the Cache class for temporarily storing events.
        health_checker (HealthChecker): An instance of the HealthChecker class for monitoring
                                        component health.

    Methods:
        subscribe_connector(event: router.SubscribeConnector): Handles subscriptions for the connector component.
        subscribe(event: router.Subscribe) -> Optional[Connection]: Subscribes a component to an event type.
        unsubscribe(event: router.Unsubscribe): Unsubscribes a component from an event type.
        unsubscribe_pipe(pipe: Connection): Removes a component's subscriptions based on its connection.
        send_from(event: router.SendFrom): Sends cached events to subscribed components based on criteria.
        send(event: Event, connections: Set[Connection]): Distributes an event to subscribed components.
        run(): Starts the event routing process, handling incoming events and managing subscriptions.
        quit(): Gracefully shuts down the router, ensuring all resources are cleaned up properly.
    """

    def __init__(self, queue: "Queue[Event]", components: Set[ManagerBase]) -> None:
        """
        Initializes the Router with a queue for incoming events and a set of components
        to be monitored for health checks.
        """
        super().__init__()
        self.queue: "Queue[Event]" = queue
        self.logger = getLogger(__name__)
        self.connections: Dict[str, Connection] = {}
        self.subscribed_requests: Dict[Type[Event], Set[Connection]] = defaultdict(set)
        self.subscribed_responses: Dict[Type[Event], Set[Connection]] = defaultdict(set)
        self.connector_pipe: Connection = None  # type: ignore
        self.cache = Cache()
        self.health_checker = HealthChecker(self.queue, components)

    def _add_subscription(self, subscriptions: Dict[Type[Event], Set[Connection]], event_type: Type[Event], pipe: Connection) -> None:
        """
        Adds a subscription for a given event type and connection.
        """
        if event_type in subscriptions:
            subscriptions[event_type].add(pipe)
        else:
            subscriptions[event_type] = {pipe}

    def _remove_subscription(self, subscriptions: Dict[Type[Event], Set[Connection]], event_type: Type[Event], pipe: Connection) -> None:
        """
        Removes a subscription for a given event type and connection.
        """
        if event_type in subscriptions:
            subscriptions[event_type].discard(pipe)

    def subscribe(self, event: router.Subscribe) -> Optional[Connection]:
        """
         Registers a component for receiving events of a specific type.
         """
        # Every serialized object is different... this should enforce consistency
        saved_pipe = self.connections.get(event.name, None)
        if saved_pipe is not None:
            event.pipe = saved_pipe
        elif event.pipe is not None:
            self.connections[event.name] = event.pipe
        else:
            self.logger.error("Missing pipe in subscribe event!")
            return None

        if event.request:
            self._add_subscription(self.subscribed_requests, event.event_type, event.pipe)
        else:
            self._add_subscription(self.subscribed_responses, event.event_type, event.pipe)

        self.health_checker.component_add(event.name, event.pipe)
        if self.connector_pipe is not None:  # If switch connector is running, send him info about new request
            subscribed_new = router.SubscribedNew(event=event.event_type, request=event.request)
            self.connector_pipe.send(subscribed_new)

        event.response.set_status(ResponseStatus.OK)
        pipe = event.pipe
        event.pipe = None
        pipe.send(event)
        self.logger.debug("Subscribed for %s event: %s", event.name, event.event_type)
        return pipe

    def subscribed_all(self, event: router.SubscribedAll):
        event.response.demands = router.Demands(
            requests=set(k for k in self.subscribed_requests.keys() if k in shareable_events),
            responses=set(k for k in self.subscribed_responses.keys() if k in shareable_events))
        self.connector_pipe = event.pipe
        event.pipe = None
        self.connector_pipe.send(event)

    def subscribe_connector(self, event: router.SubscribeConnector) -> None:
        """
        Registers the connector component for receiving specific types of events.
        """
        if not self.connector_pipe:
            self.logger.error("Connector component is not running!")
            return
        if event.request:
            self.subscribed_requests[event.event].add(self.connector_pipe)
        else:
            self.subscribed_responses[event.event].add(self.connector_pipe)

    def unsubscribe(self, event: router.Unsubscribe) -> None:
        """
        Removes a component's subscription for a specific event type.
        """
        self._remove_subscription(
            self.subscribed_requests if event.request else self.subscribed_responses,
            event.event_type,
            event.pipe
        )

    def unsubscribe_pipe(self, pipe: Connection) -> None:
        """
        Removes all subscriptions associated with a given connection.
        """
        self.logger.warning("Removing subscription for this connection: %s", pipe)
        for manager, connection in self.connections.items():
            if connection == pipe:
                del self.connections[manager]
                break

        for event, subscribers in self.subscribed_requests.items():
            self.subscribed_requests[event] = {subscriber for subscriber in subscribers if subscriber != pipe}

        for event, subscribers in self.subscribed_responses.items():
            self.subscribed_responses[event] = {subscriber for subscriber in subscribers if subscriber != pipe}

    def last_events(self, event: LastEvents) -> None:
        event.response.last_events = self.cache.last_events()
        self.connector_pipe.send(event)

    def send_from(self, event: router.SendFrom) -> None:
        """
        Sends events from the cache to subscribed components based on specified criteria.
        """
        if event.key:
            cached_events = self.cache.events_from(event.key)
        else:
            cached_events = [item.event for item in self.cache.items]
        self.logger.debug("Sending cached events: %s", len(cached_events))
        for cache_event in cached_events:
            if not cache_event.has_response():
                if self.connector_pipe in self.subscribed_requests[cache_event.__class__]:
                    self.logger.debug("Sending to switch: %s", cache_event)
                    event.response.event = cache_event
                    self.connector_pipe.send(event)
            else:
                if self.connector_pipe in self.subscribed_responses[cache_event.__class__]:
                    self.logger.debug("Sending to switch: %s", cache_event)
                    event.response.event = cache_event
                    self.connector_pipe.send(event)

    def send(self, event: Event, connections: Set[Connection]) -> None:
        """
        Distributes an event to all subscribed components.
        """
        time = self.cache.append(event)
        if not connections:
            self.logger.info("Event without subscriber: %s", event)
            return

        for connection in connections:
            if event._switch and event._switch.received and self.connector_pipe == connection:  # This supposes to stop event storm
                self.logger.debug("Event discarded: %s", event)
                continue  # pragma: no cover
            try:
                connection.send(event)
            except BrokenPipeError as e:
                self.logger.error("Broken pipe - %s", connection, exc_info=e)
                self.unsubscribe_pipe(connection)
                self.health_checker.restart_component(connection, time)

    def run(self) -> None:
        """
        Continuously processes incoming events, routing them according to subscriptions
        and managing system health checks.
        """
        self.logger.info("Router has started!")
        while True:
            event: Event = self.queue.get()
            self.logger.debug("<- %s", event)
            if not event.has_response():
                if isinstance(event, group.Router):
                    if isinstance(event, router.Subscribe):
                        pipe = self.subscribe(event)
                        if not pipe:
                            continue
                        for undelivered_event in self.cache.find(event.event_type, True, event.from_time):
                            self.logger.debug("Sending cached events: %s", undelivered_event)
                            try:
                                pipe.send(undelivered_event)
                            except BrokenPipeError as e:
                                self.logger.error("Broken pipe - %s", pipe, exc_info=e)
                                self.unsubscribe_pipe(pipe)
                    elif isinstance(event, router.Unsubscribe):
                        self.unsubscribe(event)
                    elif isinstance(event, router.SubscribedAll):
                        self.subscribed_all(event)
                    elif isinstance(event, router.SubscribeConnector):
                        self.subscribe_connector(event)
                    elif isinstance(event, router.LastEvents):
                        self.last_events(event)
                    elif isinstance(event, router.SendFrom):
                        self.send_from(event)
                    elif isinstance(event, router.HealthCheck):
                        self.health_checker.control_start()
                        self.send(event, self.subscribed_requests.get(event.__class__, set()))
                    else:
                        self.logger.info("Unknown event: %s", event)
                else:
                    self.send(event, self.subscribed_requests[event.__class__])
            else:
                if isinstance(event, group.Router):
                    if isinstance(event, router.HealthCheck):
                        if not event.response.name:
                            self.logger.error("Wrong health check response - missing 'name'")
                            continue
                        self.health_checker.control_result(event.response.name, event.response.get_status())
                    else:
                        self.logger.info("Unknown event with response: %s", event)
                else:
                    self.send(event, self.subscribed_responses[event.__class__])

    def quit(self) -> None:
        """
        Stops the router and performs necessary cleanup operations.
        """
        self.cache.quit()



from multiprocessing import Queue
from uuid import UUID

from edri.api import Middleware
from edri.api.broker import Broker
from edri.api.listener import Listener
from edri.abstract import ManagerBase
from edri.config.setting import TOKEN_LENGTH
from edri.dataclass.event import Event
from edri.router import Router
from edri.router.connector import Connector
from edri.utility.manager import Store
from edri.utility.manager.scheduler import Scheduler, Job


class EDRI:
    def __init__(self) -> None:
        self.router_queue: "Queue[Event]" = Queue()
        self.components: set[ManagerBase] = set()
        self.store: Store
        self.scheduler: Scheduler
        self.api: Listener
        self.api_broker: Broker
        self.connector: Connector
        self.router: Router

    def add_component(self, component: ManagerBase) -> None:
        self.components.add(component)

    def start_store(self) -> None:
        self.store = Store(self.router_queue)
        self.store.start()

    def start_scheduler(self, jobs: list[Job] | None = None) -> None:
        if not jobs:
            jobs = []
        self.scheduler = Scheduler(self.router_queue, jobs)
        self.scheduler.start()

    def api_broker_start(self, middlewares: list[Middleware] = None) -> None:
        if middlewares is None:
            middlewares = []
        self.api_broker = Broker(self.router_queue, Queue(), middlewares)
        self.api_broker.start()

    def start_api(self, middlewares: list[Middleware] | None = None) -> None:
        if middlewares is None:
            middlewares = []
        if not hasattr(self, "api_broker"):
            self.api_broker_start(middlewares=middlewares)
        else:
            self.api_broker.middlewares = middlewares
        self.api = Listener(self.api_broker.api_broker_queue, middlewares)
        self.api.start()

    def start_connector(self, router_id: UUID | None = None) -> None:
        self.connector = Connector(self.router_queue, router_id)
        self.connector.start()

    def run(self) -> None:
        self.router = Router(self.router_queue, self.components)
        for component in self.components:
            component.start(self.router_queue)
        try:
            self.router.run()
        except KeyboardInterrupt:
            self.router.quit()
            if hasattr(self, "connector") and self.connector:
                self.connector.quit()

__all__ = ["EDRI"]


from setuptools import setup, find_packages

long_description = """
# Event Driven Routing Infrastructure (EDRI)

EDRI is a framework designed to simplify the development of parallel and distributed applications. It allows developers to efficiently run code across multiple threads, processes, or even machines by leveraging an event-driven architecture. EDRI emphasizes flexibility and scalability, enabling the development of systems that can handle complex workflows with ease.

The framework is ideal for applications that require high throughput, low latency, and the ability to scale horizontally across distributed systems. It is designed to handle dynamic, real-time data processing and communication through the use of well-defined events.

## Key Features

- **Event-Driven Architecture**: EDRI uses events as data packets to enable communication between multiple senders and recipients. Events can be notifications, requests, or tunnels, each serving a specific communication purpose.

- **Routing Infrastructure**: The exchange and delivery of events are managed similarly to TCP/IP in computer networks. The router directs events to appropriate recipients based on event types, ensuring efficient communication.

- **Scalable Architecture**: The framework supports running tasks as methods, threads, and processes, allowing for scalable and parallel execution of tasks across multiple computing nodes.

- **Flexible Components**: EDRI includes various components such as Router, Manager, Worker, Scheduler, and Key-Value Store, each designed to handle specific responsibilities within the system.

## Architecture Overview

```mermaid
graph TD;
    A[Events] --> B[Router];
    B --> C[Manager];
    C --> D[Worker];
    B --> E[Scheduler];
    B --> F[Key-Value Store];
    B --> G[API];
```

## Components

### Event

Events are data packets created using Python data classes to enable communication between multiple senders and recipients. They serve as the core communication mechanism within EDRI and can be categorized into:

- **Notifications**: Events that announce actions or changes in the system.
- **Requests**: Events that solicit a response or action from another component.
- **Tunnels**: Events that facilitate continuous data transfer between components.

Events are strictly defined to ensure clarity and consistency across the system. They can carry responses or simply announce actions, depending on their type.

### Router

The Router is the central component responsible for forwarding events between different parts of the framework, typically managers. It ensures that events are delivered to the appropriate recipients based on their event types.

Key responsibilities of the Router include:

- Directing events to the appropriate managers or workers.
- Managing connections to external systems via the Connector.
- Caching connections to handle short-term network outages without disrupting operations.

#### Connector

The Connector is a specialized part of the Router that manages the connection between the Router and external systems (referred to as the Switch). It handles:

- Bridging internal communication with external networks using TCP/IP sockets.
- Distributing routed events to other routers and vice versa.
- Ensuring continuity of events to prevent loss or duplication during connection interruptions.

### Manager

The Manager distributes tasks contained in the transmitted events to individual workers. It supports running tasks in various modes:

- **Methods**: Tasks executed as regular function calls within the same process.
- **Threads**: Tasks executed in separate threads for parallelism within the same process.
- **Processes**: Tasks executed in separate processes for true parallelism across CPU cores.

The Manager also handles data transfer for tunnels and subscribes to the event types it needs to receive.

### Scheduler

The Scheduler is a special type of Manager used for time-defined event sending. It enables scheduling tasks or events at specific times, such as:

- Performing backups during off-peak hours.
- Sending regular updates or heartbeat signals.

### Key-Value Store

The Key-Value Store is another specialized Manager that provides a built-in storage mechanism for caching and retrieving data. It is useful for:

- Caching data during processes like data downloads.
- Periodically updating stored data using the Scheduler.
- Sharing state or configuration data between components.

### Worker

Workers are classes defined by the programmer to perform specific tasks or operations. They are the execution units in the EDRI framework and can be customized to handle a wide range of functionalities.

### API

The API serves as the interface between clients and the EDRI system. It handles:

- Converting incoming client requests into events.
- Processing events through the system.
- Returning responses back to the clients.

**Features of the API include:**

- Support for dynamic content generation.
- Management of event-based interactions, including both request-type and notification-type events.
- Real-time communication capabilities between the server and clients.

## Use Cases

EDRI is suitable for applications that require:

- **High Throughput and Low Latency**: Efficient handling of a large number of events with minimal delay.
- **Real-Time Data Processing**: Immediate processing and response to incoming data.
- **Horizontal Scalability**: Ability to scale across multiple machines or nodes.
- **Complex Workflows**: Management of intricate processes that benefit from an event-driven approach.

## Getting Started

To start using EDRI:

1. **Define Your Events**: Use Python data classes to define the events that will be used for communication between components.

2. **Set Up Your Router and Managers**: Initialize the Router and any necessary Managers (including Scheduler and Key-Value Store if needed).

3. **Implement Your Workers**: Create Worker classes that perform the specific tasks required by your application.

4. **Use the API**: Utilize the API to interact with the system, handle client requests, and facilitate communication.

5. **Run and Monitor Your Application**: Execute your application and monitor the event flow to ensure everything operates as expected.

## Conclusion

EDRI provides a robust framework for developing scalable, parallel, and distributed applications using an event-driven architecture. By abstracting the complexities of inter-process and network communication, it allows developers to focus on implementing the core logic of their applications.

Whether you're building a real-time data processing system, a distributed service, or any application that benefits from parallel execution, EDRI offers the tools and structure to make development more manageable and efficient.
"""

setup(
    name='edri',
    version='2024.09.01',
    packages=find_packages(),
    description='Event Driven Routing Infrastructure',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Marek Olšan',
    author_email='marek.olsan@gmail.com',
    install_requires=[
        "typeguard>=4.0",
        "websockets>=12.0",
        "requests>=2.0",
        "validators>=0.22.0",
        "uvicorn>=0.27.0",
        "typing_extensions>=4.0",
    ],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.12",
        "Operating System :: POSIX",
        "Operating System :: POSIX :: Linux",
        "Intended Audience :: Developers",
        "Natural Language :: Czech",
        "Typing :: Typed"
    ],
)
